from ampel.core.AmpelContext import AmpelContext
from fastapi import Depends, FastAPI, status
from fastapi.encoders import jsonable_encoder
from fastapi.exceptions import HTTPException
from functools import cache
from pydantic import ValidationError

from .settings import settings
from .job import render_job, compact_json
from .auth import get_user, User
from .models import ArgoJobModel
from . import api

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    import httpx

app = FastAPI(
    title="Argo job service",
    version="0.1.0",
    root_path=settings.root_path,
)


@cache
def get_context():
    return AmpelContext.load(
        config=settings.ampel_config,
        freeze_config=True,
    )


@app.post("/jobs/lint")
async def lint_job(job: ArgoJobModel):
    """
    Check a job template for errors
    """
    try:
        return render_job(get_context(), job)
    except ValidationError as exc:
        raise HTTPException(
            status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
            detail=jsonable_encoder(exc.errors()),
        )


@app.post("/jobs")
async def submit_job(job: ArgoJobModel, user: User = Depends(get_user)):
    """
    Submit a job template
    """
    try:
        template = render_job(get_context(), job)
    except ValidationError as exc:
        raise HTTPException(
            status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
            detail=jsonable_encoder(exc.errors()),
        )
    settings = api.KubernetesSettings.get()
    async with api.api_client() as client:
        response = await client.post(
            f"api/v1/workflow-templates/{settings.namespace}",
            json={
                "template": {
                    "metadata": {
                        "name": job.name,
                        "labels": {
                            "ampelproject.github.io/creator": user.name,
                        },
                        "annotations": {
                            "ampelproject.github.io/job": compact_json(job.json()),
                        },
                    },
                    **template,
                }
            },
        )
    if response.status_code >= status.HTTP_400_BAD_REQUEST:
        raise HTTPException(status_code=response.status_code, detail=response.json())
    return response.json()

# If we are mounted under a (non-stripped) prefix path, create a potemkin root
# router and mount the actual root as a sub-application. This has no effect
# other than to prefix the paths of all routes with the root path.
if settings.root_path:
    wrapper = FastAPI()
    wrapper.mount(settings.root_path, app)
    app = wrapper


import secrets
from typing import Set

from pydantic import (
    BaseSettings,
    Field,
)


class Settings(BaseSettings):
    root_path: str = Field("", env="ROOT_PATH")
    ampel_config: str = Field("", env="AMPEL_CONFIG")
    ampel_image: str = Field("gitlab.desy.de:5555/jakob.van.santen/docker-ampel:v0.8", env="AMPEL_IMAGE")
    ampel_secrets: str = Field("ampel-secrets", env="AMPEL_SECRETS")
    image_pull_secrets: list[str] = Field(["desy-gitlab-registry"], env="IMAGE_PULL_SECRETS")

    jwt_secret_key: str = Field(secrets.token_urlsafe(64), env="JWT_SECRET_KEY")
    jwt_algorithm: str = Field("HS256", env="JWT_ALGORITHM")
    allowed_identities: Set[str] = Field(
        {"AmpelProject", "ZwickyTransientFacility"},
        env="ALLOWED_IDENTITIES",
        description="Usernames, teams, and orgs allowed to create workflow templates",
    )

    class Config:
        env_file = ".env"


settings = Settings()



from pydantic import Field

from ampel.model.job.JobModel import JobModel


class ArgoJobModel(JobModel):
    name: str = Field(
        ...,
        regex="[a-z0-9]([-a-z0-9]*[a-z0-9])?(\\.[a-z0-9]([-a-z0-9]*[a-z0-9])?)*",
        description="Name of the job template. Must be a lowercase RFC 1123 subdomain name.",
    )



from typing import Any
from ampel.config.AmpelConfig import AmpelConfig
from ampel.model.UnitModel import UnitModel
from ampel.model.job.JobModel import JobModel, TaskUnitModel, TemplateUnitModel
from ampel.core.AmpelContext import AmpelContext
from ampel.abstract.AbsProcessorTemplate import AbsProcessorTemplate
# avoid a circular import in UnitLoader._validate_unit_model
from ampel.abstract.AbsProcessController import AbsProcessController

from importlib import import_module
from pydantic import ValidationError
from contextlib import contextmanager
import json

from typing import Literal

from .settings import settings


def get_job_template(
    image="gitlab.desy.de:5555/jakob.van.santen/docker-ampel:v0.8",
    channel: list[dict[str, Any]] = [],
    alias: dict[Literal["t0", "t1", "t2", "t3"], Any] = {},
) -> dict[str, Any]:
    return {
        "name": "ampel-job",
        "inputs": {
            "parameters": [
                {"name": "task"},
                {"name": "name"},
                {"name": "url", "value": ""},
            ],
            "artifacts": [
                {
                    "name": "task",
                    "path": "/config/task.yml",
                    "raw": {"data": "{{inputs.parameters.task}}"},
                },
                {
                    "name": "channel",
                    "path": "/config/channel.yml",
                    "raw": {"data": compact_json(channel)},
                },
                {
                    "name": "alias",
                    "path": "/config/alias.yml",
                    "raw": {"data": compact_json(alias)},
                },
                {
                    "name": "alerts",
                    "path": "/data/alerts.avro",
                    "http": {"url": "{{ inputs.parameters.url }}"},
                    "optional": True,
                },
            ],
        },
        "outputs": {},
        "metadata": {},
        "container": {
            "name": "main",
            "image": image,
            "command": [
                "ampel",
                "process",
                "--config",
                "/opt/env/etc/ampel.yml",
                "--secrets",
                "/config/secrets/secrets.yaml",
                "--channel",
                "/config/channel.yml",
                "--alias",
                "/config/alias.yml",
                "--db",
                "{{workflow.parameters.db}}",
                "--schema",
                "/config/task.yml",
                "--name",
                "{{inputs.parameters.name}}",
            ],
            "env": [
                {
                    "name": "AMPEL_CONFIG_resource.mongo",
                    "valueFrom": {
                        "secretKeyRef": {
                            "name": "mongo-live-admin-superuser",
                            "key": "connectionString.standard",
                        }
                    },
                }
            ],
            "resources": {},
            "volumeMounts": [
                {
                    "name": "secrets",
                    "readOnly": True,
                    "mountPath": "/config/secrets",
                }
            ],
        },
    }


def get_unit_model(task: TaskUnitModel) -> dict[str, Any]:
    """get dict representation of UnitModel from TaskUnitModel"""
    return task.dict(exclude={"title", "multiplier"})


def render_task_template(ctx: AmpelContext, model: TemplateUnitModel) -> TaskUnitModel:
    """
    Resolve and validate a full AbsEventUnit config from given template
    """
    if model.template not in ctx.config._config["template"]:
        raise ValueError(f"Unknown process template: {model.template}")

    fqn = ctx.config._config["template"][model.template]
    class_name = fqn.split(".")[-1]
    Tpl = getattr(import_module(fqn), class_name)
    if not issubclass(Tpl, AbsProcessorTemplate):
        raise ValueError(f"Unexpected template type: {Tpl}")

    tpl = Tpl(**model.config)

    return TaskUnitModel(
        **(
            tpl.get_model(ctx.config._config, model.dict()).dict()
            | {"title": model.title, "multiplier": model.multiplier}
        )
    )


@contextmanager
def job_context(ctx: AmpelContext, job: JobModel):
    """Add custom channels and aliases defined in the job"""
    old_config = ctx.config
    try:
        config = AmpelConfig(old_config.get(), freeze=False)
        config_dict = config._config
        for c in job.channel:
            dict.__setitem__(config_dict["channel"], str(c["channel"]), c)

        for k, v in job.alias.items():
            if "alias" not in config_dict:
                dict.__setitem__(config_dict, "alias", {})
            for kk, vv in v.items():
                if k not in config_dict["alias"]:
                    dict.__setitem__(config_dict["alias"], k, {})
                dict.__setitem__(config_dict["alias"][k], kk, vv)
        config.freeze()
        ctx.config = config
        ctx.loader.config = config
        yield ctx
    finally:
        ctx.config = old_config
        ctx.loader.config = old_config


compact_json = json.JSONEncoder(separators=(",", ":")).encode


def render_job(context: AmpelContext, job: JobModel):
    """Render Ampel job into an Argo workflow template spec"""

    steps = []

    with job_context(context, job) as ctx:
        for num, task_def in enumerate(job.task):

            task = (
                render_task_template(ctx, task_def)
                if isinstance(task_def, TemplateUnitModel)
                else task_def
            )
            # always raise exceptions
            if task.override is None:
                task.override = {}
            task.override["raise_exc"] = True

            with ctx.loader.validate_unit_models():
                unit: UnitModel = UnitModel(**get_unit_model(task))

            if not "AbsEventUnit" in ctx.config._config["unit"][task.unit]["base"]:
                raise ValidationError(
                    [[ValueError(f"{task.unit} is not a subclass of AbsEventUnit")]],
                    model=JobModel,
                )

            title = task.title or f"{job.name}-{num}"

            sub_step = {
                "template": "ampel-job",
                "arguments": {
                    "parameters": [
                        {"name": "name", "value": title},
                        {"name": "task", "value": compact_json(unit.dict())},
                    ]
                },
            }

            steps.append(
                [
                    {
                        "name": title + (f"-{idx}" if task.multiplier else ""),
                    }
                    | sub_step
                    for idx in range(task.multiplier)
                ]
            )

    return {
        "spec": {
            "templates": [
                get_job_template(
                    image=settings.ampel_image,
                    channel=job.channel,
                    alias=job.alias,
                ),
                {
                    "name": "workflow",
                    "inputs": {},
                    "outputs": {},
                    "metadata": {},
                    "steps": steps,
                },
            ],
            "entrypoint": "workflow",
            "arguments": {
                "parameters": [
                    {"name": "url"},
                    {"name": "name"},
                    {"name": "db"},
                ]
            },
            "serviceAccountName": "argo-workflow",
            "volumes": [
                {"name": "secrets", "secret": {"secretName": settings.ampel_secrets}}
            ],
            "ttlStrategy": {"secondsAfterCompletion": 1200},
            "podGC": {"strategy": "OnPodCompletion"},
            "workflowMetadata": {
                "labels": {"example": "true"},
            },
            "imagePullSecrets": [{"name": n} for n in settings.image_pull_secrets],
        },
    }



# interact with the argo API

import httpx
from contextlib import asynccontextmanager
from functools import cache
from pydantic import BaseSettings, Field

from typing import AsyncIterator


class KubernetesSettings(BaseSettings):
    serviceaccount: str = Field(..., env="SERVICE_ACCOUNT_TOKEN")
    base_url: str = Field(..., env="ARGO_BASE_URL")
    namespace: str = Field("ampel", env="ARGO_NAMESPACE")

    verify_ssl: bool = Field(True, env="VERIFY_SSL")

    @classmethod
    @cache
    def get(cls) -> "KubernetesSettings":
        kwargs = {}
        try:
            with open("/var/run/secrets/kubernetes.io/serviceaccount/token") as f:
                kwargs["serviceaccount"] = f.read()
        except FileNotFoundError:
            pass
        return cls(**kwargs)


@asynccontextmanager
async def api_client() -> AsyncIterator[httpx.AsyncClient]:
    settings = KubernetesSettings.get()
    async with httpx.AsyncClient(base_url=settings.base_url, verify=settings.verify_ssl) as client:
        client.headers.update({"Authorization": f"Bearer {settings.serviceaccount}"})
        yield client


async def post_manifest(manifest: dict):
    settings = KubernetesSettings.get()
    async with api_client() as client:
        response = await client.post(
            f"api/v1/workflow-templates/{settings.namespace}", json=manifest
        )
        
        response.raise_for_status()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['ampel', 'ampel.argo']

package_data = \
{'': ['*']}

install_requires = \
['PyJWT>=2.3.0,<3.0.0',
 'ampel-core>=0.8.2-alpha.13,<0.8.3',
 'fastapi>=0.75.0,<0.76.0',
 'httpx>=0.22.0,<0.23.0',
 'uvicorn>=0.17.6,<0.18.0']

setup_kwargs = {
    'name': 'ampel-argo',
    'version': '0.1.0a4',
    'description': '',
    'long_description': None,
    'author': 'Jakob van Santen',
    'author_email': 'jvansanten@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)
